#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <ctype.h>
#include <utime.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <syslog.h>
#include <netdb.h>

void strlwr (char *dst, const char *src)
{
	while (*src != '\0'){
		*dst++ = tolower(*src);
		src++;
	}
	*dst = '\0';
}


/*
	Extract which virtual host is expected by the POP client.
	Return -1 if any error.
*/
int vhost_getourname(char *name, int maxsiz)
{
	int ret = -1;
	struct sockaddr_in adr;
	int len = sizeof(adr);
	if (getsockname (0,(struct sockaddr*)&adr,&len) != -1){
		struct hostent *ent = gethostbyaddr ((const char*)&adr.sin_addr
			,sizeof(adr.sin_addr.s_addr),AF_INET);
		syslog (LOG_DEBUG,"connexion de %p",ent);
		if (ent != NULL && strlen(ent->h_name)<maxsiz-1){
			strlwr (name,ent->h_name);
			syslog (LOG_DEBUG,"connexion de %s",name);
			ret = 0;
		}
	}
	return ret;
}

static time_t file_date(const char *fname)
{
	time_t ret = 0;
	struct stat st;
	if (stat(fname,&st)!=-1) ret = st.st_mtime;
	return ret;
}

static int file_setdate (const char *fname, time_t date)
{
	struct utimbuf timbuf;
	timbuf.modtime = date;
	timbuf.actime = date;
	return utime (fname,&timbuf);
}


static int vhost_updpass (
	const char *vpass,	// VIrtual passwd file
	const char *pass,	// normal passwd file in the virtual home
	time_t date,		// Date to apply to the file generated
	int skip)		// Fix the home directory
{
	int ret = -1;
	FILE *fin = fopen (vpass,"r");
	if (fin != NULL){
		char tmp[PATH_MAX];
		FILE *fout;
		sprintf (tmp,"%s.TMP",pass);
		fout = fopen (tmp,"w");
		if (fout != NULL){
			char buf[1000];
			while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
				char *tb[100];
				int nb = 1;
				char *pt = buf;
				tb[0] = buf;
				while ((pt = strchr(pt,':'))!=NULL){
					*pt++ = '\0';
					tb[nb++] = pt;
				}
				if (nb > 5){
					const char *ctl = "%s";
					int i;
					tb[5] += skip;
					for (i=0; i<nb; i++){
						fprintf (fout,ctl,tb[i]);
						ctl = ":%s";
					}
				}
			}
			fclose (fout);
			file_setdate (tmp,date);
			ret = rename (tmp,pass);
		}
	}
	return ret;
}


/*
	Copy the passwd file from /etc/vmail/passwd.domain to
	/vhome/domain/etc/passwd if the date are different

	Do the same for the shadow file if exist
*/
static int vhost_setuppasswd (const char *domain)
{
	int ret = -1;
	char vpasswd[PATH_MAX],vshadow[PATH_MAX];
	char passwd[PATH_MAX],shadow[PATH_MAX];
	char vhome[PATH_MAX];
	time_t vdate,date;
	sprintf (vhome,"/vhome/%s",domain);
	sprintf (vpasswd,"/etc/vmail/passwd.%s",domain);
	sprintf (vshadow,"/etc/vmail/shadow.%s",domain);
	sprintf (passwd,"%s/etc/passwd",vhome);
	sprintf (shadow,"%s/etc/shadow",vhome);
	vdate = file_date (vpasswd);
	date = file_date (passwd);
	if (vdate != date){
		// Update is needed
		syslog (LOG_NOTICE,"Updating %s",passwd);
		ret = vhost_updpass (vpasswd,passwd,vdate,strlen(vhome));
		vhost_updpass (vshadow,shadow,vdate,0);
		// We don't care if there is no shadow
		chmod (shadow,0600);
	}
	return ret;
}

static int vhost_finddomain (char *domain)
{
	int ret = -1;
	struct stat st;
	char path[PATH_MAX];
	sprintf (path,"/vhome/%s",domain);
	if (stat(path,&st)!=-1
		&& S_ISDIR(st.st_mode)){
		ret = 0;
	}else{
		char *pt = strchr(domain,'.');
		if (pt != NULL){
			pt++;
			sprintf (path,"/vhome/%s",pt);
			if (stat(path,&st)!=-1
			&& S_ISDIR(st.st_mode)){
				char tmp[PATH_MAX];
				strcpy (tmp,pt);
				strcpy (domain,tmp);
				ret = 0;
			}
		}
	}
	return ret;
}


int main (int argc, char *argv[])
{
	char domain[PATH_MAX];
	openlog ("vhost",LOG_PID,LOG_DAEMON);
	if (vhost_getourname(domain,sizeof(domain))!=-1
		&& vhost_finddomain (domain)!=-1){
		char path[PATH_MAX];
		char pathetc[PATH_MAX];
		sprintf (path,"/vhome/%s",domain);
		sprintf (pathetc,"%s/etc",path);
		if (file_date(pathetc)!=0){
			vhost_setuppasswd (domain);
			syslog (LOG_NOTICE,"Changing directory/root to %s",path);
			chdir (path);
			chroot (path);
		}else{
			syslog (LOG_NOTICE,"No /etc directory for vdomain %s: using main domain"
				,domain);
		}
	}
	execv (argv[1],argv+1);	
	return 1;
}

